;;; ns-menu-bar.el --- define a default menu bar for NS users

;; Author: Carl Edman
;; Keywords: internal
;; Maintainer: Adrian Robert <arobert@cogsci.ucsd.edu>

;; Copyright (C) 1993 Free Software Foundation, Inc.

;; This file is part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Code:

(require 'cl)
(require 'ispell)

(require 'menu-bar)
;; restore the menu-bar from before menu-bar.el
(if (featurep 'menu-bar)
    (progn
      (define-key global-map [menu-bar] nil)
      (makunbound 'menu-bar-edit-menu)
      (makunbound 'menu-bar-tools-menu)
      (makunbound 'menu-bar-file-menu)
      (makunbound 'yank-menu)
      (and (= yank-menu-length 20)
           (makunbound 'yank-menu-length))
      (fmakunbound 'clipboard-yank)
      (fmakunbound 'clipboard-kill-ring-save)
      (fmakunbound 'clipboard-kill-region)
      (fmakunbound 'menu-bar-enable-clipboard)
      ))
      
;; Don't clobber an existing menu-bar keymap, to preserve any menu-bar key
;; definitions made in loaddefs.el.
(or (lookup-key global-map [menu-bar])
    (define-key global-map [menu-bar] (make-sparse-keymap "menu-bar")))

;; Utility function
(defun add-to-keymap-end (keymap key def)
  (let* ((contents (reverse (cdr keymap)))
	 (name (car contents))
	 (revmap (append (list 'keymap) (cdr contents) (list name))))
    (define-key revmap key def)
    (append (list 'keymap) (cdr (reverse (cdr revmap))) (list name))))


(cond ((eq system-type 'darwin)
       (setq menu-bar-final-items '(buffer windows services help-menu)))
      ;; otherwise, gnustep
      (t
       (setq menu-bar-final-items '(buffer windows services hide-app quit)) )
)

(cond ((not (eq system-type 'darwin))  ;; i.e., gnustep
      (define-key global-map [menu-bar quit] '("Quit" . save-buffers-kill-emacs))
      (define-key global-map [menu-bar hide-app] '("Hide" . do-hide-emacs))
))

(define-key global-map [menu-bar services] (cons "Services" (make-sparse-keymap "Services")))
(define-key global-map [menu-bar windows] (make-sparse-keymap "Windows"))
(define-key global-map [menu-bar buffer] (cons "Buffers" (make-sparse-keymap "Buffers")))
;(define-key global-map [menu-bar buffers] (make-sparse-keymap "Buffers"))

(defvar menu-bar-tools-menu (make-sparse-keymap "Tools"))
(define-key global-map [menu-bar tools] (cons "Tools" menu-bar-tools-menu))

(define-key global-map [menu-bar options] (cons "Options" menu-bar-options-menu))

(defvar menu-bar-edit-menu (make-sparse-keymap "Edit"))
(define-key global-map [menu-bar edit] (cons "Edit" menu-bar-edit-menu))

(defvar menu-bar-file-menu (make-sparse-keymap "File"))
(define-key global-map [menu-bar file] (cons "File" menu-bar-file-menu))

;; If running under GNUstep, rename "Help" to "Info"
(cond ((eq system-type 'darwin)
       (define-key global-map [menu-bar help-menu]
	 (cons "Help" menu-bar-help-menu)))
      (t
       (let ((contents (reverse (cdr menu-bar-help-menu))))
	 (setq menu-bar-help-menu
	       (append (list 'keymap) (cdr contents) (list "Info"))))
       (define-key global-map [menu-bar help-menu]
	 (cons "Info" menu-bar-help-menu))))



(define-key menu-bar-tools-menu [games]
  (list 'menu-item "Games" menu-bar-games-menu))

(define-key menu-bar-tools-menu [separator-games]
  '("--"))

(define-key menu-bar-tools-menu [simple-calculator]
  '("Simple Calculator" . calculator))
(define-key menu-bar-tools-menu [calc]
  '("Programmable Calculator" . calc))
(define-key menu-bar-tools-menu [calendar]
  '("Calendar" . calendar))

(define-key menu-bar-tools-menu [separator-net]
  '("--"))

(define-key menu-bar-tools-menu [directory-search]
  '("Directory Search" . eudc-tools-menu))
(define-key menu-bar-tools-menu [compose-mail]
  '("Send Mail" . compose-mail))
(define-key menu-bar-tools-menu [rmail]
  '("Read Mail" . menu-bar-read-mail))

(define-key menu-bar-tools-menu [gnus] '("Read Net News (Gnus)" . gnus))

(define-key menu-bar-tools-menu [separator-vc] '("--"))

(define-key menu-bar-tools-menu [vc]
  (cons "Version Control" vc-menu-map))

(define-key menu-bar-tools-menu [separator-vc] '("--"))
(define-key menu-bar-tools-menu [pcl-cvs] '("PCL-CVS" . cvs-global-menu))

(define-key menu-bar-tools-menu [ediff-misc]
  '("Ediff Miscellanea" . menu-bar-ediff-misc-menu))
(define-key menu-bar-tools-menu [epatch]
  (cons "Apply Patch" menu-bar-epatch-menu))
(define-key menu-bar-tools-menu [ediff-merge]
  (cons "Merge" menu-bar-ediff-merge-menu))
(define-key menu-bar-tools-menu [compare]
  (cons "Compare (Ediff)" menu-bar-ediff-menu))

(define-key menu-bar-tools-menu [separator-spell]
  '("--"))

; Note, this is also under Edit menu for *Step consistency, but putting
; it here too makes it easier for people used to X-based Emacs to find it
(define-key menu-bar-tools-menu [spell] '("Spell Checking" . ispell-menu-map))

(define-key menu-bar-tools-menu [separator-prog] '("--"))

(define-key menu-bar-tools-menu [gdb]
  '("Debugger (GDB)..." . gdb))
(define-key menu-bar-tools-menu [shell-on-region]
  '("Shell Command on Region..." . shell-command-on-region))
(define-key menu-bar-tools-menu [shell]
  '("Shell Command..." shell-command))
(define-key menu-bar-tools-menu [compile]
  '("Compile..." . compile))
(define-key menu-bar-tools-menu [grep]
  '("Search Files (with grep)..." . grep))


(define-key menu-bar-file-menu [one-window]
  '("Remove Splits" . delete-other-windows))
(define-key menu-bar-file-menu [split-window]
  '("Split Window" . split-window-vertically))

(define-key menu-bar-file-menu [separator-print] '("--"))

(defvar ps-print-menu-map (make-sparse-keymap "Postscript Print"))
(define-key ps-print-menu-map [ps-print-region]
  '("Region (B+W)" . ps-print-region))
(define-key ps-print-menu-map [ps-print-buffer]
  '("Buffer (B+W)" . ps-print-buffer))
(define-key ps-print-menu-map [ps-print-region-faces]
  '("Region" . ps-print-region-with-faces))
(define-key ps-print-menu-map [ps-print-buffer-faces]
  '("Buffer" . ps-print-buffer-with-faces))
(define-key menu-bar-file-menu [postscript-print]
  (cons "Postscript Print" ps-print-menu-map))

(define-key menu-bar-file-menu [print-region]
  '("Print Region" . print-region))
(define-key menu-bar-file-menu [print-buffer]
  '("Print Buffer" . print-buffer))

(define-key menu-bar-file-menu [separator-save] '("--"))

(define-key menu-bar-file-menu [recover-session]
  '("Recover Crashed Session" . recover-session))
(define-key menu-bar-file-menu [revert-buffer]
  '("Revert Buffer" . revert-buffer))
(define-key menu-bar-file-menu [write-file]
  '("Save Buffer As..." . ns-write-file-using-panel))
(define-key menu-bar-file-menu [save-buffer] '("Save Buffer" . save-buffer))

(define-key menu-bar-file-menu [kill-buffer]
  '("Kill Current Buffer" . kill-this-buffer))
(define-key menu-bar-file-menu [delete-this-frame]
  '("Close Frame" . delete-frame))

(define-key menu-bar-file-menu [separator-open] '("--"))

(define-key menu-bar-file-menu [insert-file]
  '("Insert File..." . insert-file))
(define-key menu-bar-file-menu [dired]
  '("Open Directory..." . ns-open-file-using-panel))
(define-key menu-bar-file-menu [open-file]
  '("Open File..." . ns-open-file-using-panel))
(define-key menu-bar-file-menu [make-frame]
  '("New Frame" . make-frame))



(define-key menu-bar-edit-menu [props] (cons "Text Properties" facemenu-menu))
(if (fboundp 'start-process)
    (define-key menu-bar-edit-menu [spell] '("Spell" . ispell-menu-map)))
(define-key menu-bar-edit-menu [fill] '("Fill" . fill-region))

(define-key menu-bar-edit-menu [separator-fill] '("--"))

(define-key menu-bar-edit-menu [bookmark]
  (cons "Bookmarks" menu-bar-bookmark-map))
(define-key menu-bar-edit-menu [goto]
  (list 'menu-item "Go To" menu-bar-goto-menu))
(define-key menu-bar-edit-menu [replace]
  (list 'menu-item "Replace" menu-bar-replace-menu))
(define-key menu-bar-edit-menu [search]
  (list 'menu-item "Search" menu-bar-search-menu))

(define-key menu-bar-edit-menu [separator-cutpaste] '("--"))

(define-key menu-bar-edit-menu [mark-whole-buffer]
  '("Select All" . mark-whole-buffer))
(define-key menu-bar-edit-menu [clear] '("Clear" . delete-region))
(defvar yank-menu (make-sparse-keymap "Select and Paste"))
(define-key menu-bar-edit-menu [select-paste]
  (cons "Select and Paste" yank-menu))
(define-key menu-bar-edit-menu [paste] '("Paste" . yank))
(define-key menu-bar-edit-menu [copy] '("Copy" . ns-copy-including-secondary))
;menu-bar-kill-ring-save
(define-key menu-bar-edit-menu [cut] '("Cut" . kill-region))

(define-key menu-bar-edit-menu [separator-undo] '("--"))

(define-key menu-bar-edit-menu [undo] '("Undo" . undo))



(defun info-ns-emacs ()         ; for the menubar as well
  "Jump to ns-emacs info item."
  (interactive)
  (info "ns-emacs"))

(define-key menu-bar-help-menu [ns-bug-report]
  '("Report Emacs.app bug..." . ns-submit-bug-report))
(define-key menu-bar-help-menu [info-ns]
  '("Emacs.app Manual" . info-ns-emacs))
(if (not (eq system-type 'darwin))
    ;; in OS X it's in the app menu already
    (define-key menu-bar-help-menu [info-panel]
      '("About Emacs..." . do-emacs-info-panel)))


(defcustom yank-menu-length 40
  "*Maximum length to display in the yank-menu."
  :type 'integer
  :group 'mouse)

(defun menu-bar-update-yank-menu (string old)
  (let* ((front (car (cdr yank-menu)))
	 ;; build menu-string
	 ;; replace newlines by dots
	 ;; and shorten the string to yank-menu-length
	 (menu-string (if (<= (length string) (* 4 yank-menu-length))
			  string
			(concat
			 (substring string 0 (* 2 yank-menu-length))
			 (substring string (- (* 2 yank-menu-length))))))
	 (menu-string (let ((res menu-string) (pos -1))
			(while (setq pos (string-match "\n" res (1+ pos)))
			  (setq res (replace-match "." nil t res)))
			(if (string-match "\\.*$" res)
			    (setq res (replace-match "" nil t res)))))
         (menu-string (if (<= (length menu-string) yank-menu-length)
                          menu-string
                        (concat
                         (substring menu-string 0 (/ yank-menu-length 2))
                         "..."
                         (substring menu-string (- (/ yank-menu-length 2)))))))
    ;; Don't let the menu string be all dashes
    ;; because that has a special meaning in a menu.
    (if (string-match "\\`-+\\'" menu-string)
	(setq menu-string (concat menu-string " ")))
    ;; If we're supposed to be extending an existing string, and that
    ;; string really is at the front of the menu, then update it in place.
    (if (and old (or (eq old (car front))
		     (string= old (car front))))
	(progn
	  (setcar front string)
	  (setcar (cdr front) menu-string))
      (setcdr yank-menu
	      (cons
	       (cons string (cons menu-string 'menu-bar-select-yank))
	       (cdr yank-menu)))))
  (if (> (length (cdr yank-menu)) 30) ; was kill-ring-max
      (setcdr (nthcdr 30 yank-menu) '("Select and Paste"))))
;;      (setcdr (nthcdr 30 yank-menu) nil)))
(defun menu-bar-select-yank ()
  (interactive "*")
  (push-mark (point))
  (insert last-command-event))

(defun menu-bar-select-frame ()
  (interactive)
  (make-frame-visible last-command-event)
  (raise-frame last-command-event)
  (select-frame last-command-event))

(defun menu-bar-update-frames ()
  ;; If user discard the Windows item, play along.
  (and (lookup-key (current-global-map) [menu-bar windows])
       (let ((frames (frame-list))
             (frames-menu (make-sparse-keymap "Select Frame")))
         (setcdr frames-menu
                 (nconc
                  (mapcar '(lambda (frame)
                             (nconc (list frame
                                          (cdr (assq 'name (frame-parameters frame)))
                                          (cons nil nil))
                                    'menu-bar-select-frame))
                          frames)
                  (cdr frames-menu)))
	 (define-key frames-menu [separator-frames] '("--"))
         (define-key frames-menu [popup-color-panel]
           '("Colors..." . ns-popup-color-panel))
         (define-key frames-menu [popup-font-panel]
           '("Font Panel..." . ns-popup-font-panel))
	 (define-key frames-menu [separator-arrange] '("--"))
         (define-key frames-menu [arrange-all-frames]
           '("Arrange All Frames" . ns-arrange-all-frames))
         (define-key frames-menu [arrange-visible-frames]
           '("Arrange Visible Frames" . ns-arrange-visible-frames))
         ;; Don't use delete-frame as event name
         ;; because that is a special event.
         (define-key (current-global-map) [menu-bar windows]
           (cons "Windows" frames-menu)))))

(defun force-menu-bar-update-buffers ()
  ;; This is a hack to get around fact that we already checked
  ;; frame-or-buffer-changed-p and reset it, so menu-bar-update-buffers
  ;; does not pick up any change.
  (menu-bar-update-buffers t))

(add-hook 'menu-bar-update-fab-hook 'menu-bar-update-frames)
(add-hook 'menu-bar-update-fab-hook 'force-menu-bar-update-buffers)

(defun menu-bar-update-frames-and-buffers ()
  (if (frame-or-buffer-changed-p)
      (run-hooks 'menu-bar-update-fab-hook)))

(setq menu-bar-update-hook
      (delq 'menu-bar-update-buffers menu-bar-update-hook))
(add-hook 'menu-bar-update-hook 'menu-bar-update-frames-and-buffers)

(menu-bar-update-frames-and-buffers)

;; ns-arrange functions contributed
;; by Eberhard Mandler <mandler@dbag.ulm.DaimlerBenz.COM>
(defun ns-arrange-all-frames ()
  "Arranges all frames according to topline"
  (interactive)
  (ns-arrange-frames t))

(defun ns-arrange-visible-frames ()
  "Arranges all visible frames according to topline"
  (interactive)
  (ns-arrange-frames nil))

(defun ns-arrange-frames ( vis)
  (let ((frame (next-frame))
	(end-frame (selected-frame))
	(inc-x 20)                      ;relative position of frames
	(inc-y 22)
	(x-pos 100)                     ;start position
	(y-pos 40)
	(done nil))
    (while (not done)                   ;cycle through all frames
      (if (not (or vis (eq (frame-visible-p frame) t)))
	(setq x-pos x-pos); do nothing; true case
	(set-frame-position frame x-pos y-pos)
	(setq x-pos (+ x-pos inc-x))
	(setq y-pos (+ y-pos inc-y))
	(raise-frame frame))
      (select-frame frame)
      (setq frame (next-frame))
      (setq done (equal frame end-frame)))
    (set-frame-position end-frame x-pos y-pos)
    (raise-frame frame)
    (select-frame frame)))


;;; Set up a menu bar menu for the minibuffer.
(mapcar
 (function
  (lambda (map)
    (define-key map [menu-bar minibuf]
      (cons "Minibuf" (make-sparse-keymap "Minibuf")))))
 (list minibuffer-local-ns-map
       minibuffer-local-must-match-map
       minibuffer-local-isearch-map
       minibuffer-local-map
       minibuffer-local-completion-map))

(mapcar
 (function
  (lambda (map)
    (define-key map [menu-bar minibuf ?\?]
      '("List Completions" . minibuffer-completion-help))
    (define-key map [menu-bar minibuf space]
      '("Complete Word" . minibuffer-complete-word))
    (define-key map [menu-bar minibuf tab]
      '("Complete" . minibuffer-complete))
    ))
 (list minibuffer-local-must-match-map
       minibuffer-local-completion-map))

(mapcar
 (function
  (lambda (map)
    (define-key map [menu-bar minibuf quit]
      '("Quit" . keyboard-escape-quit))
    (define-key map [menu-bar minibuf return]
      '("Enter" . exit-minibuffer))
    ))
 (list minibuffer-local-ns-map
       minibuffer-local-must-match-map
       minibuffer-local-isearch-map
       minibuffer-local-map
       minibuffer-local-completion-map))

(defconst ns-proto-call-service
  '((arg)
    (interactive "p")
    (let* ((in-string (if mark-active
                          (buffer-substring (region-beginning) (region-end))))
           (out-string (ns-perform-service service in-string)))
      (cond
       ((and out-string (or (not in-string)
                            (not (string= in-string out-string))))
        (if mark-active (delete-region (region-beginning) (region-end)))
        (insert-string out-string)
        (setq deactivate-mark nil))))))
  
(defun ns-define-service (path)
  (let ((mapping [menu-bar services])
        (service (mapconcat 'identity path "/"))
        (name (intern
               (mapconcat '(lambda (s) (if (= s 32) "-" (char-to-string s)))
                          (mapconcat 'identity (cons "ns-service" path) "-")
                          ""))))
    (eval (append (list 'defun name)
                  (subst service 'service ns-proto-call-service)))
    (cond
     ((lookup-key global-map mapping)
      (while (cdr path)
        (setq mapping (vconcat mapping (list (intern (car path)))))
        (if (not (keymapp (lookup-key global-map mapping)))
            (define-key global-map mapping
              (cons (car path) (make-sparse-keymap (car path)))))
        (setq path (cdr path)))
      (setq mapping (vconcat mapping (list (intern (car path)))))
      (define-key global-map mapping (cons (car path) name))))
    name))

(provide 'ns-menu-bar)

;;;; ns-menu-bar.el ends here
