;;; ns-win.el --- lisp side of interface with
;;;               NeXT/Open/GNUstep/MacOS X window system
;; Copyright (C) 1993, 1994, 2005, 2006 Free Software Foundation, Inc.

;; Author: Carl Edman, Christian Limpach, Scott Bender, Christophe de Dinechin,
;;         Adrian Robert
;; Keywords: terminals

;;; This file is part of GNU Emacs.
;;;
;;; GNU Emacs is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2, or (at your option)
;;; any later version.
;;;
;;; GNU Emacs is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with GNU Emacs; see the file COPYING.  If not, write to
;;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:

;; ns-win.el:  this file is loaded from ../lisp/startup.el when it recognizes
;; that NS windows are to be used.  Command line switches are parsed and those
;; pertaining to NS are processed and removed from the command line.  The
;; NS display is opened and hooks are set for popping up the initial window.

;; startup.el will then examine startup files, and eventually call the hooks
;; which create the first window (s).

;; A number of other NS convenience functions are defined in this file,
;; which works in close coordination with src/nsfns.m.

;;; Code:


(if (not (eq window-system 'ns))
    (error "%s: Loading ns-win.el but not compiled for *Step/OS X"
	   (invocation-name)))

(require 'frame)
(require 'mouse)
(require 'faces)
(require 'delsel)
(require 'mldrag)
(require 'ns-menu-bar)


(defconst ns-version "9.0-rc2")
;; remember to change version number in man/ns-emacs.texi and Info.plist files
(defun ns-submit-bug-report ()
   "Submit via mail a bug report on Emacs 23.0.0 for GNUstep / OS X."
   (interactive)
   (let ((frame-parameters (frame-parameters))
         (server-vendor (ns-server-vendor))
         (server-version (ns-server-version)))
     (reporter-submit-bug-report
      "Adrian Robert <Adrian.B.Robert@gmail.com>"
      ;;"Christophe de Dinechin <descubes@earthlink.net>"
      ;;"Scott Bender <emacs@harmony-ds.com>" 
      ;;"Christian Limpach <chris@nice.ch>"
      ;;"Carl Edman <cedman@princeton.edu>"
      (concat "Emacs for GNUstep / OS X " ns-version)
      '(ns-expand-space ns-cursor-blink-rate ns-alternate-modifier
        data-directory frame-parameters window-system window-system-version
        server-vendor server-version system-configuration-options))))


;;;; Command line argument handling.

(defvar ns-invocation-args nil)
(defvar ns-command-line-resources nil)

;; Handler for switches of the form "-switch value" or "-switch".
(defun ns-handle-switch (switch)
  (let ((aelt (assoc switch command-line-ns-option-alist)))
    (if aelt
	(let ((param (nth 3 aelt))
	      (value (nth 4 aelt)))
	  (if value
	      (setq default-frame-alist
		    (cons (cons param value)
			  default-frame-alist))
	    (setq default-frame-alist
		  (cons (cons param
			      (car ns-invocation-args))
			default-frame-alist)
		  ns-invocation-args (cdr ns-invocation-args)))))))

;; Handler for switches of the form "-switch n"
(defun ns-handle-numeric-switch (switch)
  (let ((aelt (assoc switch command-line-ns-option-alist)))
    (if aelt
	(let ((param (nth 3 aelt)))
	  (setq default-frame-alist
		(cons (cons param
			    (string-to-number (car ns-invocation-args)))
		      default-frame-alist)
		ns-invocation-args
		(cdr ns-invocation-args))))))

;; Make -iconic apply only to the initial frame!
(defun ns-handle-iconic (switch)
  (setq initial-frame-alist
        (cons '(visibility . icon) initial-frame-alist)))

;; Handle the -name option, set the name of
;; the initial frame.
(defun ns-handle-name-switch (switch)
  (or (consp ns-invocation-args)
      (error "%s: missing argument to `%s' option" (invocation-name) switch))
  (setq initial-frame-alist (cons (cons 'name (car ns-invocation-args))
                                  initial-frame-alist)
        ns-invocation-args (cdr ns-invocation-args)))

(defun ns-handle-nxopen (switch)
  (setq unread-command-events (append unread-command-events '(ns-open-file))
        ns-input-file (append ns-input-file (list (car ns-invocation-args)))
        ns-invocation-args (cdr ns-invocation-args)))

(defun ns-handle-nxopentemp (switch)
  (setq unread-command-events (append unread-command-events '(ns-open-temp-file))
        ns-input-file (append ns-input-file (list (car ns-invocation-args)))
        ns-invocation-args (cdr ns-invocation-args)))

(defun ns-ignore-0-arg (switch)
  )
(defun ns-ignore-1-arg (switch)
  (setq ns-invocation-args (cdr ns-invocation-args)))
(defun ns-ignore-2-arg (switch)
  (setq ns-invocation-args (cddr ns-invocation-args)))

(defun ns-handle-args (args)
  "Here the NS-related command line options in ARGS are processed,
before the user's startup file is loaded.  They are copied to
`ns-invocation-args', from which the NS related things are extracted, first
the switch (e.g., \"-fg\") in the following code, and possible values
\(e.g., \"black\") in the option handler code (e.g., ns-handle-switch).
This function returns ARGS minus the arguments that have been processed."
  ;; We use ARGS to accumulate the args that we don't handle here, to return.
  (setq ns-invocation-args args
        args nil)
  (while ns-invocation-args
    (let* ((this-switch (car ns-invocation-args))
	   (orig-this-switch this-switch)
	   completion argval aelt handler)
      (setq ns-invocation-args (cdr ns-invocation-args))
      ;; Check for long options with attached arguments
      ;; and separate out the attached option argument into argval.
      (if (string-match "^--[^=]*=" this-switch)
	  (setq argval (substring this-switch (match-end 0))
		this-switch (substring this-switch 0 (1- (match-end 0)))))
      ;; Complete names of long options.
      (if (string-match "^--" this-switch)
	  (progn
	    (setq completion (try-completion this-switch
                                             command-line-ns-option-alist))
	    (if (eq completion t)
		;; Exact match for long option.
		nil
	      (if (stringp completion)
		  (let ((elt (assoc completion command-line-ns-option-alist)))
		    ;; Check for abbreviated long option.
		    (or elt
			(error "Option `%s' is ambiguous" this-switch))
		    (setq this-switch completion))))))
      (setq aelt (assoc this-switch command-line-ns-option-alist))
      (if aelt (setq handler (nth 2 aelt)))
      (if handler
	  (if argval
	      (let ((ns-invocation-args
		     (cons argval ns-invocation-args)))
		(funcall handler this-switch))
	    (funcall handler this-switch))
	(setq args (cons orig-this-switch args)))))
  (nreverse args))

(defun ns-parse-geometry (geom)
  "Parse an NS-style geometry string STRING.
Returns an alist of the form ((top . TOP), (left . LEFT) ... ).
The properties returned may include `top', `left', `height', and `width'."
  (if (string-match "\\([0-9]+\\)\\( \\([0-9]+\\)\\( \\([0-9]+\\)\\( \\([0-9]+\\) ?\\)?\\)?\\)?"
                    geom)
      (apply 'append
	   (list
	    (list (cons 'top (string-to-number (match-string 1 geom))))
	    (if (match-string 3 geom)
		(list (cons 'left (string-to-number (match-string 3 geom)))))
	    (if (match-string 5 geom)
		(list (cons 'height (string-to-number (match-string 5 geom)))))
	    (if (match-string 7 geom)
		(list (cons 'width (string-to-number (match-string 7 geom)))))))
    '()))
(defalias 'x-parse-geometry 'ns-parse-geometry)



;;;; Keyboard mapping.

;; These tell read-char how to convert
;; these special chars to ASCII.
(put 'backspace 'ascii-character 127)
(put 'delete 'ascii-character 127)
(put 'tab 'ascii-character ?\t)
(put 'S-tab 'ascii-character (logior 16 ?\t))
(put 'linefeed 'ascii-character ?\n)
(put 'clear 'ascii-character 12)
(put 'return 'ascii-character 13)
(put 'escape 'ascii-character ?\e)

;; Map certain keypad keys into ASCII characters
;; that people usually expect.
(define-key function-key-map [backspace] [127])
(define-key function-key-map [delete] [127])
(define-key function-key-map [tab] [?\t])
(define-key function-key-map [S-tab] [25])
(define-key function-key-map [linefeed] [?\n])
(define-key function-key-map [clear] [11])
(define-key function-key-map [return] [13])
(define-key function-key-map [escape] [?\e])
(define-key function-key-map [M-backspace] [?\M-\d])
(define-key function-key-map [M-delete] [?\M-\d])
(define-key function-key-map [M-tab] [?\M-\t])
(define-key function-key-map [M-linefeed] [?\M-\n])
(define-key function-key-map [M-clear] [?\M-\013])
(define-key function-key-map [M-return] [?\M-\015])
(define-key function-key-map [M-escape] [?\M-\e])


;; Here are some NeXTSTEP like bindings for command key sequences.
(define-key global-map [?\s-,] 'ns-popup-prefs-panel)
(define-key global-map [?\s-'] 'next-multiframe-window)
(define-key global-map [?\s-`] 'other-frame)
(define-key global-map [?\s--] 'center-line)
(define-key global-map [?\s-:] 'ispell)
(define-key global-map [?\s-\;] 'ispell-next)
(define-key global-map [?\s-?] 'info)
(define-key global-map [?\s-^] 'kill-some-buffers)
(define-key global-map [?\s-&] 'kill-this-buffer)
(define-key global-map [?\s-C] 'ns-popup-color-panel)
(define-key global-map [?\s-D] 'dired)
(define-key global-map [?\s-E] 'edit-abbrevs)
(define-key global-map [?\s-L] 'shell-command)
(define-key global-map [?\s-M] 'manual-entry)
(define-key global-map [?\s-S] 'ns-write-file-using-panel)
(define-key global-map [?\s-a] 'mark-whole-buffer)
;(define-key global-map [?\s-c] 'kill-ring-save)
(define-key global-map [?\s-c] 'ns-copy-including-secondary)
(define-key global-map [?\s-d] 'isearch-repeat-backward)
(define-key global-map [?\s-e] 'isearch-yank-kill)
(define-key global-map [?\s-f] 'isearch-forward)
(define-key global-map [?\s-g] 'isearch-repeat-forward)
(define-key global-map [?\s-h] 'do-hide-emacs)
(define-key global-map [?\s-H] 'do-hide-others)
(define-key global-map [?\s-j] 'exchange-point-and-mark)
(define-key global-map [?\s-k] 'kill-this-buffer)
(define-key global-map [?\s-l] 'goto-line)
(define-key global-map [?\s-m] 'iconify-frame)
(define-key global-map [?\s-n] 'make-frame)
(define-key global-map [?\s-o] 'ns-open-file-using-panel)
(define-key global-map [?\s-p] 'print-buffer)
(define-key global-map [?\s-q] 'save-buffers-kill-emacs)
(define-key global-map [?\s-s] 'save-buffer)
(define-key global-map [?\s-t] 'ns-popup-font-panel)
(define-key global-map [?\s-u] 'revert-buffer)
(define-key global-map [?\s-v] 'yank)
(define-key global-map [?\s-w] 'delete-frame)
(define-key global-map [?\s-x] 'kill-region)
(define-key global-map [?\s-y] 'ns-paste-secondary)
(define-key global-map [?\s-z] 'undo)
(define-key global-map [?\s-|] 'shell-command-on-region)
(define-key global-map [s-kp-bar] 'shell-command-on-region)
; (as in Terminal.app)
(define-key global-map [s-right] 'ns-next-frame)
(define-key global-map [s-left] 'ns-prev-frame)

(define-key global-map [home] 'beginning-of-buffer)
(define-key global-map [end] 'end-of-buffer)
(define-key global-map [kp-home] 'beginning-of-buffer)
(define-key global-map [kp-end] 'end-of-buffer)
(define-key global-map [kp-prior] 'scroll-down)
(define-key global-map [kp-next] 'scroll-up)


;; Special NeXTSTEP generated events are converted to function keys.  Here
;; are the bindings for them.
(define-key global-map [ns-power-off]
  '(lambda () (interactive) (save-buffers-kill-emacs t)))
(define-key global-map [ns-open-file] 'ns-find-file)
(define-key global-map [ns-open-temp-file] [ns-open-file])
(define-key global-map [ns-drag-file] 'ns-insert-file)
(define-key global-map [ns-drag-color] 'ns-set-foreground-at-mouse)
(define-key global-map [S-ns-drag-color] 'ns-set-background-at-mouse)
(define-key global-map [ns-drag-text] 'ns-insert-text)
(define-key global-map [ns-change-font] 'ns-respond-to-change-font)
(define-key global-map [ns-open-file-line] 'ns-open-file-select-line)
(define-key global-map [ns-insert-working-text] 'ns-insert-working-text)
(define-key global-map [ns-delete-working-text] 'ns-delete-working-text)

(setq system-key-alist
      (list
       (cons (logior (lsh 0 16)   1) 'ns-power-off)
       (cons (logior (lsh 0 16)   2) 'ns-open-file)
       (cons (logior (lsh 0 16)   3) 'ns-open-temp-file)
       (cons (logior (lsh 0 16)   4) 'ns-drag-file)
       (cons (logior (lsh 0 16)   5) 'ns-drag-color)
       (cons (logior (lsh 0 16)   6) 'ns-drag-text)
       (cons (logior (lsh 0 16)   7) 'ns-change-font)
       (cons (logior (lsh 0 16)   8) 'ns-open-file-line)
       (cons (logior (lsh 0 16)   9) 'ns-insert-working-text)
       (cons (logior (lsh 0 16)  10) 'ns-delete-working-text)
       (cons (logior (lsh 1 16)  32) 'f1)
       (cons (logior (lsh 1 16)  33) 'f2)
       (cons (logior (lsh 1 16)  34) 'f3)
       (cons (logior (lsh 1 16)  35) 'f4)
       (cons (logior (lsh 1 16)  36) 'f5)
       (cons (logior (lsh 1 16)  37) 'f6)
       (cons (logior (lsh 1 16)  38) 'f7)
       (cons (logior (lsh 1 16)  39) 'f8)
       (cons (logior (lsh 1 16)  40) 'f9)
       (cons (logior (lsh 1 16)  41) 'f10)
       (cons (logior (lsh 1 16)  42) 'f11)
       (cons (logior (lsh 1 16)  43) 'f12)
       (cons (logior (lsh 1 16)  44) 'kp-insert)
       (cons (logior (lsh 1 16)  45) 'kp-delete)
       (cons (logior (lsh 1 16)  46) 'kp-home)
       (cons (logior (lsh 1 16)  47) 'kp-end)
       (cons (logior (lsh 1 16)  48) 'kp-prior)
       (cons (logior (lsh 1 16)  49) 'kp-next)
       (cons (logior (lsh 1 16)  50) 'print-screen)
       (cons (logior (lsh 1 16)  51) 'scroll-lock)
       (cons (logior (lsh 1 16)  52) 'pause)
       (cons (logior (lsh 1 16)  53) 'system)
       (cons (logior (lsh 1 16)  54) 'break)
       (cons (logior (lsh 1 16)  56) 'please-tell-carl-what-this-key-is-called-56)
       (cons (logior (lsh 1 16)  61) 'please-tell-carl-what-this-key-is-called-61)
       (cons (logior (lsh 1 16)  62) 'please-tell-carl-what-this-key-is-called-62)
       (cons (logior (lsh 1 16)  63) 'please-tell-carl-what-this-key-is-called-63)
       (cons (logior (lsh 1 16)  64) 'please-tell-carl-what-this-key-is-called-64)
       (cons (logior (lsh 1 16)  69) 'please-tell-carl-what-this-key-is-called-69)
       (cons (logior (lsh 1 16)  70) 'please-tell-carl-what-this-key-is-called-70)
       (cons (logior (lsh 1 16)  71) 'please-tell-carl-what-this-key-is-called-71)
       (cons (logior (lsh 1 16)  72) 'please-tell-carl-what-this-key-is-called-72)
       (cons (logior (lsh 1 16)  73) 'please-tell-carl-what-this-key-is-called-73)
       (cons (logior (lsh 2 16)   3) 'kp-enter)
       (cons (logior (lsh 2 16)   9) 'kp-tab)
       (cons (logior (lsh 2 16)  28) 'kp-quit)
       (cons (logior (lsh 2 16)  35) 'kp-hash)
       (cons (logior (lsh 2 16)  42) 'kp-multiply)
       (cons (logior (lsh 2 16)  43) 'kp-add)
       (cons (logior (lsh 2 16)  44) 'kp-separator)
       (cons (logior (lsh 2 16)  45) 'kp-subtract)
       (cons (logior (lsh 2 16)  46) 'kp-decimal)
       (cons (logior (lsh 2 16)  47) 'kp-divide)
       (cons (logior (lsh 2 16)  48) 'kp-0)
       (cons (logior (lsh 2 16)  49) 'kp-1)
       (cons (logior (lsh 2 16)  50) 'kp-2)
       (cons (logior (lsh 2 16)  51) 'kp-3)
       (cons (logior (lsh 2 16)  52) 'kp-4)
       (cons (logior (lsh 2 16)  53) 'kp-5)
       (cons (logior (lsh 2 16)  54) 'kp-6)
       (cons (logior (lsh 2 16)  55) 'kp-7)
       (cons (logior (lsh 2 16)  56) 'kp-8)
       (cons (logior (lsh 2 16)  57) 'kp-9)
       (cons (logior (lsh 2 16)  60) 'kp-less)
       (cons (logior (lsh 2 16)  61) 'kp-equal)
       (cons (logior (lsh 2 16)  62) 'kp-more)
       (cons (logior (lsh 2 16)  64) 'kp-at)
       (cons (logior (lsh 2 16)  92) 'kp-backslash)
       (cons (logior (lsh 2 16)  96) 'kp-backtick)
       (cons (logior (lsh 2 16) 124) 'kp-bar)
       (cons (logior (lsh 2 16) 126) 'kp-tilde)
       (cons (logior (lsh 2 16) 157) 'kp-mu)
       (cons (logior (lsh 2 16) 165) 'kp-yen)
       (cons (logior (lsh 2 16) 167) 'kp-paragraph)
       (cons (logior (lsh 2 16) 172) 'left)
       (cons (logior (lsh 2 16) 173) 'up)
       (cons (logior (lsh 2 16) 174) 'right)
       (cons (logior (lsh 2 16) 175) 'down)
       (cons (logior (lsh 2 16) 176) 'kp-ring)
       (cons (logior (lsh 2 16) 201) 'kp-square)
       (cons (logior (lsh 2 16) 204) 'kp-cube)
       (cons (logior (lsh 3 16)   8) 'backspace)
       (cons (logior (lsh 3 16)   9) 'tab)
       (cons (logior (lsh 3 16)  10) 'linefeed)
       (cons (logior (lsh 3 16)  11) 'clear)
       (cons (logior (lsh 3 16)  13) 'return)
       (cons (logior (lsh 3 16)  18) 'pause)
       (cons (logior (lsh 3 16)  25) 'S-tab)
       (cons (logior (lsh 3 16)  27) 'escape)
       (cons (logior (lsh 3 16) 127) 'delete)
       ))

;; alt-up/down scrolling a la Stuart.app
(defun up-one () (interactive) (scroll-up 1))
(defun down-one () (interactive) (scroll-down 1))
(global-set-key [M-up] 'down-one)
(global-set-key [M-down] 'up-one)

(defun left-one () (interactive) (scroll-left 1))
(defun right-one () (interactive) (scroll-right 1))
; These conflict w/word-left, word-right
;;(global-set-key [M-left] 'left-one)
;;(global-set-key [M-right] 'right-one)



;;;; Miscellaneous mouse bindings.

;; Mouse bindings for mode line drag
;(global-set-key [mode-line down-mouse-1] 'mldrag-drag-mode-line)
;(global-set-key [vertical-line down-mouse-1] 'mldrag-drag-vertical-line)

;;; Allow shift-clicks to work just like under NS
(defun mouse-extend-region (event)
  "Move point or mark so as to extend region.
This should be bound to a mouse click event type."
  (interactive "e")
  (mouse-minibuffer-check event)
  (let ((posn (event-end event)))
    (if (not (windowp (posn-window posn)))
        (error "Cursor not in text area of window"))
    (select-window (posn-window posn))
    (cond
     ((not (numberp (posn-point posn))))
     ((or (not mark-active) (> (abs (- (posn-point posn) (point)))
                               (abs (- (posn-point posn) (mark)))))
      (let ((point-save (point)))
        (unwind-protect
            (progn
              (goto-char (posn-point posn))
              (push-mark nil t t)
              (or transient-mark-mode
                  (sit-for 1)))
          (goto-char point-save))))
     (t
      (goto-char (posn-point posn))))))

(define-key global-map [S-mouse-1] 'mouse-extend-region)
(global-unset-key [S-down-mouse-1])


;;;; Composed key sequence handling.

(defface ns-working-text-face
  '((t :underline t))
  "Face used to highlight working text during compose sequence insert."
  :group 'ns)

(defvar ns-working-overlay nil
  "Overlay used to highlight working text during compose sequence insert.")
(make-variable-buffer-local 'ns-working-overlay)
(defvar ns-working-overlay-len 0
  "Length of working text during compose sequence insert.")
(make-variable-buffer-local 'ns-working-overlay-len)

; Based on mac-win.el 2007/08/26 unicode-2.
(defun ns-in-echo-area ()
  "Whether, for purposes of inserting working composition text, the minibuffer
is currently being used."
  (or isearch-mode
      (and cursor-in-echo-area (current-message))
      ;; Overlay strings are not shown in some cases.
      (get-char-property (point) 'invisible)
      (and (not (bobp))
	   (or (and (get-char-property (point) 'display)
		    (eq (get-char-property (1- (point)) 'display)
			(get-char-property (point) 'display)))
	       (and (get-char-property (point) 'composition)
		    (eq (get-char-property (1- (point)) 'composition)
			(get-char-property (point) 'composition)))))))

(defun ns-insert-working-text ()
  "Insert contents of ns-working-text as UTF8 string and mark with
ns-working-overlay.  Any previously existing working text is cleared first.
The overlay is assigned the face ns-working-text-face."
  (interactive)
  (if ns-working-overlay (ns-delete-working-text))
  (let ((start (point)))
    (insert ns-working-text)
    (overlay-put (setq ns-working-overlay (make-overlay start (point)
							(current-buffer) nil t))
		 'face 'ns-working-text-face)
    (setq ns-working-overlay-len (+ ns-working-overlay-len (- (point) start)))))

(defun ns-echo-working-text ()
  "Echo contents of ns-working-text in message display area.
See ns-insert-working-text."
  (if ns-working-overlay (ns-unecho-working-text))
  (let* ((msg (current-message))
	 (msglen (length msg))
	 message-log-max)
    (setq ns-working-overlay-len (length ns-working-text))
    (setq msg (concat msg ns-working-text))
    (put-text-property msglen (+ msglen ns-working-overlay-len) 'face 'ns-working-text-face msg)
    (message "%s" msg)
    (setq ns-working-overlay t)))

(defun ns-delete-working-text()
  "Delete working text and clear ns-working-overlay."
  (interactive)
  (delete-backward-char ns-working-overlay-len)
  (setq ns-working-overlay-len 0)
  (delete-overlay ns-working-overlay))

(defun ns-unecho-working-text()
  "Delete working text from echo area and clear ns-working-overlay."
  (let ((msg (current-message))
	message-log-max)
    (setq msg (substring msg 0 (- (length msg) ns-working-overlay-len)))
    (setq ns-working-overlay-len 0)
    (setq ns-working-overlay nil)))


;;;; OS X file system Unicode UTF-8 NFD (decomposed form) support
;; Lisp code based on utf-8m.el, by Seiji Zenitani, Eiji Honjoh, and
;; Carsten Bormann.
(if (eq system-type 'darwin)
    (progn

      (defun ns-utf8-nfd-post-read-conversion (length)
	"Calls ns-convert-utf8-nfd-to-nfc to compose char sequences."
	(save-excursion
	  (save-restriction
	    (narrow-to-region (point) (+ (point) length))
	    (let ((str (buffer-string)))
	      (delete-region (point-min) (point-max))
	      (insert (ns-convert-utf8-nfd-to-nfc str))
	      (- (point-max) (point-min))
	      ))))

      (define-coding-system 'utf-8-nfd
	"UTF-8 NFD (decomposed) encoding."
	:coding-type 'utf-8
	:mnemonic ?U
	:charset-list '(unicode)
	:post-read-conversion 'ns-utf8-nfd-post-read-conversion)
      (set-file-name-coding-system 'utf-8-nfd)))



;;;; Inter-app communications support.

(defun ns-insert-text ()
  "Insert contents of ns-input-text at point."
  (interactive)
  (insert ns-input-text)
  (setq ns-input-text nil))
 
(defun ns-insert-file ()
  "Insert contents of file ns-input-file like insert-file but with less
prompting.  If file is a directory perform a find-file on it."
  (interactive)
  (let ((f))
    (setq f (car ns-input-file))
    (setq ns-input-file (cdr ns-input-file))
    (if (file-directory-p f)
        (find-file f)
      (push-mark (+ (point) (car (cdr (insert-file-contents f))))))))

(defvar ns-select-overlay nil
  "Overlay used to highlight areas in files requested by NS apps.")
(make-variable-buffer-local 'ns-select-overlay)

(defun ns-open-file-select-line ()
  "Brings up a buffer containing file ns-input-file,\n\
and highlights lines indicated by ns-input-line."
  (interactive)
  (ns-find-file)
  (cond
   ((and ns-input-line (buffer-modified-p))
    (if ns-select-overlay
        (setq ns-select-overlay (delete-overlay ns-select-overlay)))
    (deactivate-mark)
    (goto-line (if (consp ns-input-line)
                   (min (car ns-input-line) (cdr ns-input-line))
                 ns-input-line)))
   (ns-input-line
    (if (not ns-select-overlay)
        (overlay-put (setq ns-select-overlay (make-overlay (point-min) (point-min)))
                     'face 'highlight))
    (let ((beg (save-excursion
                 (goto-line (if (consp ns-input-line)
                                (min (car ns-input-line) (cdr ns-input-line))
                              ns-input-line))
                 (point)))
          (end (save-excursion
                 (goto-line (+ 1 (if (consp ns-input-line)
                                     (max (car ns-input-line) (cdr ns-input-line))
                                   ns-input-line)))
                 (point))))
      (move-overlay ns-select-overlay beg end)
      (deactivate-mark)
      (goto-char beg)))
   (t
    (if ns-select-overlay
        (setq ns-select-overlay (delete-overlay ns-select-overlay))))))

(defun ns-unselect-line ()
  "Removes any NS highlight a buffer may contain."
  (if ns-select-overlay
      (setq ns-select-overlay (delete-overlay ns-select-overlay))))

(add-hook 'first-change-hook 'ns-unselect-line)



;;;; Preferences handling.

(defun get-lisp-resource (arg1 arg2)
  (let ((res (ns-get-resource arg1 arg2)))
    (cond
     ((not res) 'unbound)
     ((string-equal (upcase res) "YES") t)
     ((string-equal (upcase res) "NO")  nil)
     (t (read res)))))

(defun ns-save-preferences ()
  "Set all the defaults."
  (interactive)
  ;; Global preferences
  (ns-set-resource nil "AlternateModifier" (symbol-name ns-alternate-modifier))
  (ns-set-resource nil "CommandModifier" (symbol-name ns-command-modifier))
  (ns-set-resource nil "ControlModifier" (symbol-name ns-control-modifier))
  (ns-set-resource nil "FunctionModifier" (symbol-name ns-function-modifier))
  (ns-set-resource nil "CursorBlinkRate"
                (if ns-cursor-blink-rate
                    (number-to-string ns-cursor-blink-rate)
                  "NO"))
  (ns-set-resource nil "ExpandSpace"
                (if ns-expand-space
                    (number-to-string ns-expand-space)
                  "NO"))
  (ns-set-resource nil "GSFontAntiAlias" (if ns-antialias-text "YES" "NO"))
  (ns-set-resource nil "UseQuickdrawSmoothing"
		   (if ns-use-qd-smoothing "YES" "NO"))
  (ns-set-resource nil "UseSystemHighlightColor"
		   (if ns-use-system-highlight-color "YES" "NO"))
  ;; Default frame parameters
  (let ((p (frame-parameters)))
    (let ((f (assq 'font p)))
      (if f (ns-set-resource nil "Font" (ns-font-name (cdr f)))))
    (let ((fs (assq 'fontsize p)))
      (if fs (ns-set-resource nil "FontSize" (number-to-string (cdr fs)))))
    (let ((fgc (assq 'foreground-color p)))
      (if fgc (ns-set-resource nil "Foreground" (cdr fgc))))
    (let ((bgc (assq 'background-color p)))
      (if bgc (ns-set-resource nil "Background" (cdr bgc))))
    (let ((cc (assq 'cursor-color p)))
      (if cc (ns-set-resource nil "CursorColor" (cdr cc))))
    (let ((ct (assq 'cursor-type p)))
      (if ct (ns-set-resource nil "CursorType"
                  (if (symbolp (cdr ct)) (symbol-name (cdr ct)) (cdr ct)))))
    (let ((under (assq 'underline p)))
      (if under (ns-set-resource nil "Underline"
				 (cond ((eq (cdr under) t) "YES")
				       ((eq (cdr under) nil) "NO")
				       (t (cdr under))))))
    (let ((ibw (assq 'internal-border-width p)))
      (if ibw (ns-set-resource nil "InternalBorderWidth"
                            (number-to-string (cdr ibw)))))
     (let ((vsb (assq 'vertical-scroll-bars p)))
       (if vsb (ns-set-resource nil "VerticalScrollBars" (cond
                                       ((eq t (cdr vsb)) "YES")
                                       ((eq nil (cdr vsb)) "NO")
                                       ((eq 'left (cdr vsb)) "left")
                                       ((eq 'right (cdr vsb)) "right")
                                       (t nil)))))
    (let ((height (assq 'height p)))
      (if height (ns-set-resource nil "Height"
                               (number-to-string (cdr height)))))
    (let ((width (assq 'width p)))
      (if width (ns-set-resource nil "Width"
                              (number-to-string (cdr width)))))
    (let ((top (assq 'top p)))
      (if top (ns-set-resource nil "Top"
                            (number-to-string (cdr top)))))
    (let ((left (assq 'left p)))
      (if left (ns-set-resource nil "Left"
                             (number-to-string (cdr left)))))
    ;; These not fully supported
    (let ((ar (assq 'auto-raise p)))
      (if ar (ns-set-resource nil "AutoRaise"
                            (if (cdr ar) "YES" "NO"))))
    (let ((al (assq 'auto-lower p)))
      (if al (ns-set-resource nil "AutoLower"
                            (if (cdr al) "YES" "NO"))))
    (let ((mbl (assq 'menu-bar-lines p)))
      (if mbl (ns-set-resource nil "Menus"
                            (if (cdr mbl) "YES" "NO"))))
    )
  (let ((fl (face-list)))
    (while (consp fl)
      (or (eq 'default (car fl))
          ;; dont save Default* since it causes all created faces to
          ;; inherit its values.  The properties of the default face
          ;; have already been saved from the frame-parameters anyway.
          (let* ((name (symbol-name (car fl)))
                 (font (face-font (car fl)))
;                 (fontsize (face-fontsize (car fl)))
                 (foreground (face-foreground (car fl)))
                 (background (face-background (car fl)))
                 (underline (face-underline-p (car fl)))
                 (italic (face-italic-p (car fl)))
                 (bold (face-bold-p (car fl)))
                 (stipple (face-stipple (car fl))))
;            (ns-set-resource nil (concat name ".attributeFont")
;                          (if font font nil))
;            (ns-set-resource nil (concat name ".attributeFontSize")
;                          (if fontsize (number-to-string fontsize) nil))
            (ns-set-resource nil (concat name ".attributeForeground")
                          (if foreground foreground nil))
            (ns-set-resource nil (concat name ".attributeBackground")
                          (if background background nil))
            (ns-set-resource nil (concat name ".attributeUnderline")
                          (if underline "YES" nil))
            (ns-set-resource nil (concat name ".attributeItalic")
                          (if italic "YES" nil))
            (ns-set-resource nil (concat name ".attributeBold")
                          (if bold "YES" nil))
            (and stipple
                 (or (stringp stipple)
                     (setq stipple (prin1-to-string stipple))))
            (ns-set-resource nil (concat name ".attributeStipple")
                          (if stipple stipple nil))))
      (setq fl (cdr fl)))))

;; call ns-save-preferences when menu-bar-options-save is called
(fset 'menu-bar-options-save-orig (symbol-function 'menu-bar-options-save))
(defun ns-save-options ()
  (interactive)
  (menu-bar-options-save-orig)
  (ns-save-preferences))
(fset 'menu-bar-options-save (symbol-function 'ns-save-options))


;;;; File handling.

(defun ns-open-file-using-panel ()
  "Pop up open-file panel, and load the result in a buffer."
  (interactive)
  ; prompt dir defaultName isLoad initial
  (setq ns-input-file (ns-read-file-name "Select File to Load" nil t nil))
  (if ns-input-file
      (and (setq ns-input-file (list ns-input-file)) (ns-find-file))))

(defun ns-write-file-using-panel ()
  "Pop up save-file panel, and save buffer in resulting name."
  (interactive)
  (let (ns-output-file)
    ; prompt dir defaultName isLoad initial
    (setq ns-output-file (ns-read-file-name "Save As" nil nil nil))
    (message ns-output-file)
    (if ns-output-file (write-file ns-output-file))))

(defun ns-find-file ()
  "Do a find-file with the ns-input-file as argument."
  (interactive)
  (let ((f) (file) (bufwin1) (bufwin2))
    (setq f (file-truename (car ns-input-file)))
    (setq ns-input-file (cdr ns-input-file))
    (setq file (find-file-noselect f))
    (setq bufwin1 (get-buffer-window file 'visible))
    (setq bufwin2 (get-buffer-window "*scratch*" 'visibile))
    (cond
     (bufwin1
      (select-frame (window-frame bufwin1))
      (raise-frame (window-frame bufwin1))
      (select-window bufwin1))
     ((and (eq ns-pop-up-frames 'fresh) bufwin2)
      (ns-hide-emacs 'activate)
      (select-frame (window-frame bufwin2))
      (raise-frame (window-frame bufwin2))
      (select-window bufwin2)
      (find-file f))
     (ns-pop-up-frames
      (ns-hide-emacs 'activate)
      (let ((pop-up-frames t)) (pop-to-buffer file nil)))
     (t
      (ns-hide-emacs 'activate)
      (find-file f)))))



;;;; Frame-related functions.

;; Don't show the frame name; that's redundant with NS.
(setq-default mode-line-frame-identification '("  "))

(defvar ns-pop-up-frames 'fresh
  "* Should file opened upon request from the Workspace be opened in a new frame ?
If t, always.  If nil, never.  Otherwise a new frame is opened
unless the current buffer is a scratch buffer.")

;; You say tomAYto, I say tomAHto..
(defvaralias 'ns-option-modifier 'ns-alternate-modifier)

(defun do-hide-emacs ()
  (interactive)
  (ns-hide-emacs t))

(defun do-hide-others ()
  (interactive)
  (ns-hide-others))

(defun do-emacs-info-panel ()
  (interactive)
  (ns-emacs-info-panel))

(defun iconify-or-deiconify-frame ()
  "Iconify the selected frame, or deiconify if it's currently an icon."
  (interactive)
  (if (eq (cdr (assq 'visibility (frame-parameters))) t)
      (iconify-frame)
    (make-frame-visible)))

(defun ns-next-frame ()
  "Switch to next visible frame."
  (interactive)
  (other-frame 1))
(defun ns-prev-frame ()
  "Switch to previous visible frame."
  (interactive)
  (other-frame -1))

; If no position specified, make new frame offset by 25 from current.
(add-hook 'before-make-frame-hook
          '(lambda ()
             (let ((left (cdr (assq 'left (frame-parameters))))
                   (top (cdr (assq 'top (frame-parameters)))))
	       (if (consp left) (setq left (cadr left)))
	       (if (consp top) (setq top (cadr top)))
               (cond
                ((or (assq 'top parameters) (assq 'left parameters)))
                ((or (not left) (not top)))
                (t
                 (setq parameters (cons (cons 'left (+ left 25))
                                        (cons (cons 'top (+ top 25))
                                              parameters))))))))

; frame will be focused anyway, so select it
(add-hook 'after-make-frame-functions 'select-frame)

(defun ns-win-suspend-error ()
  (error "Suspending an emacs running under *Step/OS X makes no sense"))
(add-hook 'suspend-hook 'ns-win-suspend-error)
(substitute-key-definition 'suspend-emacs 'iconify-or-deiconify-frame
			   global-map)

;; Based on a function by David Reitter <dreitter@inf.ed.ac.uk> ;
;; see http://lists.gnu.org/archive/html/emacs-devel/2005-09/msg00681.html .
(defun ns-toggle-toolbar (&optional frame)
  "Switches the tool bar on and off in frame FRAME.
 If FRAME is nil, the change applies to the selected frame."
  (interactive)
  (modify-frame-parameters frame
	   (list (cons 'tool-bar-lines
		       (if (> (or (frame-parameter frame 'tool-bar-lines) 0) 0)
				   0 1)) ))
  (if (not tool-bar-mode) (tool-bar-mode t)))

; Redefine from frame.el
(define-minor-mode blink-cursor-mode
  "Toggle blinking cursor mode.
With a numeric argument, turn blinking cursor mode on if ARG is positive,
otherwise turn it off.  When blinking cursor mode is enabled, the
cursor of the selected window blinks.

Note that this command is effective only when Emacs
displays through a window system, because then Emacs does its own
cursor display.  On a text-only terminal, this is not implemented."
  :init-value (not (or noninteractive
		       no-blinking-cursor
		       (eq system-type 'ms-dos)
		       (not (memq window-system '(x w32 mac)))))
  :initialize 'custom-initialize-safe-default
  :group 'cursor
  :global t
  (if blink-cursor-mode
      (setq ns-cursor-blink-rate '0.5)
      (setq ns-cursor-blink-rate nil)))


;;;; Font support.

(defalias 'x-list-fonts 'ns-list-fonts)
;; Needed for font listing functions under both backend and normal
(setq scalable-fonts-allowed t)

;; Set to use font panel instead
(defalias 'generate-fontset-menu 'ns-popup-font-panel)
(defalias 'mouse-set-font 'ns-popup-font-panel)

(defun ns-respond-to-change-font ()
  "Respond to changeFont: event, expecting ns-input-font and\n\
ns-input-fontsize of new font."
  (interactive)
  (modify-frame-parameters (selected-frame)
                           (list (cons 'font ns-input-font)
                                 (cons 'fontsize ns-input-fontsize)))
  (set-frame-font ns-input-font))


;; New font back-end support.

(require 'fontset)

;; Default fontset for Mac OS X.  This is mainly here to show how a fontset
;; can be set up manually.  Ordinarily, fontsets are auto-created whenever
;; a font is chosen by 
(defvar ns-standard-fontset-spec
; Only some code supports this so far, so use uglier XLFD version
; "-ns-*-*-*-*-*-10-*-*-*-*-*-fontset-standard,latin:Courier,han:Kai"
"-ns-*-*-*-*-*-10-*-*-*-*-*-fontset-standard,latin:-*-Courier-*-*-*-*-10-*-*-*-*-*-iso10646-1,han:-*-Kai-*-*-*-*-10-*-*-*-*-*-iso10646-1,cyrillic:-*-Trebuchet$MS-*-*-*-*-10-*-*-*-*-*-iso10646-1"
 "String of fontset spec of the standard fontset.
This defines a fontset consisting of the Courier and other fonts that
come with OS X\".

See the documentation of `create-fontset-from-fontset-spec for the format.")

;; Conditional on new-fontset so bootstrapping works on non-GUI compiles
(if (fboundp 'new-fontset)
    (progn
      ;; Setup the default fontset.
      (setup-default-fontset)
      ;; Create the standard fontset.
      (create-fontset-from-fontset-spec ns-standard-fontset-spec t)
))

;(setq default-frame-alist (cons (cons 'font "-ns-*-*-*-*-*-10-*-*-*-*-*-fontset-standard") default-frame-alist))


;;;; Pasteboard support.

(defun ns-get-pasteboard ()
  "Returns the value of the pasteboard."
  (ns-get-cut-buffer-internal 'PRIMARY))

(defun ns-set-pasteboard (string)
  "Store STRING into the NS server's pasteboard."
  ;; Check the data type of STRING.
  (substring string 0 0)
  (ns-store-cut-buffer-internal 'PRIMARY string))

;;; We keep track of the last text selected here, so we can check the
;;; current selection against it, and avoid passing back our own text
;;; from ns-pasteboard-value.
(defvar ns-last-selected-text nil)

;;; Put TEXT, a string, on the pasteboard.
(defun ns-select-text (text &optional push)
  ;; Don't send the pasteboard too much text.
  ;; It becomes slow, and if really big it causes errors.
  (ns-set-pasteboard text)
  (setq ns-last-selected-text text))

;;; Return the value of the current NS selection.  For compatibility
;;; with older NS applications, this checks cut buffer 0 before
;;; retrieving the value of the primary selection.
(defun ns-pasteboard-value ()
  (let (text)
    
    ;; Consult the selection, then the cut buffer.  Treat empty strings
    ;; as if they were unset.
    (or text (setq text (ns-get-pasteboard)))
    (if (string= text "") (setq text nil))
    
    (cond
     ((not text) nil)
     ((eq text ns-last-selected-text) nil)
     ((string= text ns-last-selected-text)
      ;; Record the newer string, so subsequent calls can use the `eq' test.
      (setq ns-last-selected-text text)
      nil)
     (t
      (setq ns-last-selected-text text)))))

(defun ns-copy-including-secondary ()
  (interactive)
  (call-interactively 'kill-ring-save)
  (ns-store-cut-buffer-internal 'SECONDARY
				(buffer-substring (point) (mark t))))
(defun ns-paste-secondary ()
  (interactive)
  (insert (ns-get-cut-buffer-internal 'SECONDARY)))

(setq interprogram-cut-function 'ns-select-text
      interprogram-paste-function 'ns-pasteboard-value)
;      interprogram-paste-function 'cut-buffer-or-selection-value)

; Some code can't or won't go through proper channels...
(defalias 'x-disown-selection-internal 'ns-disown-selection-internal)
(defalias 'x-get-selection-internal 'ns-get-selection-internal)
(defalias 'x-own-selection-internal 'ns-own-selection-internal)

(set-face-background 'region "ns_selection_color")



;;;; Scrollbar handling.

(global-set-key [vertical-scroll-bar down-mouse-1] 'ns-handle-scroll-bar-event)
(global-unset-key [vertical-scroll-bar mouse-1])
(global-unset-key [vertical-scroll-bar drag-mouse-1])

(defun ns-scroll-bar-move (event)
  "Scroll the frame according to an NS scroller event."
  (interactive "e")
  (let* ((pos (event-end event))
         (window (nth 0 pos))
         (scale (nth 2 pos)))
    (save-excursion
      (set-buffer (window-buffer window))
      (cond
       ((eq (car scale) (cdr scale))
	(goto-char (point-max)))
       ((= (car scale) 0)
	(goto-char (point-min)))
       (t
	(goto-char (+ (point-min) 1
		      (scroll-bar-scale scale (- (point-max) (point-min)))))))
      (beginning-of-line)
      (set-window-start window (point))
      (vertical-motion (/ (window-height window) 2) window))))

(defun ns-handle-scroll-bar-event (event)
  "Handle scroll bar EVENT to emulate Mac Toolbox style scrolling."
  (interactive "e")
  (let* ((position (event-start event))
	 (bar-part (nth 4 position))
	 (window (nth 0 position))
	 (old-window (selected-window)))
    (cond
     ((eq bar-part 'ratio)
      (ns-scroll-bar-move event))
     ((eq bar-part 'handle)
      (if (eq window (selected-window))
	  (track-mouse (ns-scroll-bar-move event))
	; track-mouse faster for selected window, slower for unselected
	(ns-scroll-bar-move event)))
     (t
      (select-window window)
      (cond
       ((eq bar-part 'up)
	(goto-char (window-start window))
	(scroll-down 1))
       ((eq bar-part 'above-handle)
	(pager-page-up))
       ((eq bar-part 'below-handle)
	(pager-page-down))
       ((eq bar-part 'down)
	(goto-char (window-start window))
	(scroll-up 1)))
      (select-window old-window)))))


;;;; Color support.

(defvar x-colors (ns-list-colors)
  "The list of colors defined in non-PANTONE color files.")
(defvar colors x-colors
  "The list of colors defined in non-PANTONE color files.")

(defun ns-defined-colors (&optional frame)
  "Return a list of colors supported for a particular frame.
The argument FRAME specifies which frame to try.
The value may be different for frames on different NS displays."
  (or frame (setq frame (selected-frame)))
  (let ((all-colors x-colors)
	(this-color nil)
	(defined-colors nil))
    (while all-colors
      (setq this-color (car all-colors)
	    all-colors (cdr all-colors))
;      (and (face-color-supported-p frame this-color t)
      (setq defined-colors (cons this-color defined-colors)))
;)
    defined-colors))
(defalias 'x-defined-colors 'ns-defined-colors)
(defalias 'xw-defined-colors 'ns-defined-colors)
;(defalias 'x-display-color-cells 'ns-display-color-cells)
;(defalias 'x-display-planes 'ns-display-planes)

;; Convenience and work-around for fact that set color fns now require named.
(defun ns-set-background-alpha (alpha)
  "Sets alpha (opacity) of background.
Set from 0.0 (fully transparent) to 1.0 (fully opaque; default).
Note, tranparency works better on Tiger (10.4) and higher."
  (interactive "nSet background alpha to: ")
  (let ((bgcolor (cdr (assq 'background-color (frame-parameters)))))
    (set-frame-parameter (selected-frame)
			 'background-color (ns-set-alpha bgcolor alpha))))

;; Functions for color panel + drag
(defun ns-face-at-pos (pos)
  (let* ((frame (car pos))
         (frame-pos (cons (cadr pos) (cddr pos)))
         (window (window-at (car frame-pos) (cdr frame-pos) frame))
         (window-pos (coordinates-in-window-p frame-pos window))
         (buffer (window-buffer window))
         (edges (window-edges window)))
    (cond
     ((not window-pos)
      nil)
     ((eq window-pos 'mode-line)
      'modeline)
     ((eq window-pos 'vertical-line)
      'default)
     ((consp window-pos)
      (save-excursion
        (set-buffer buffer)
        (let ((p (car (compute-motion (window-start window)
                                      (cons (nth 0 edges) (nth 1 edges))
                                      (window-end window)
                                      frame-pos
                                      (- (window-width window) 1)
                                      nil
                                      window))))
          (cond
           ((eq p (window-point window))
            'cursor)
           ((and mark-active (< (region-beginning) p) (< p (region-end)))
            'region)
           (t
	    (let ((faces (get-char-property p 'face window)))
	      (if (consp faces) (car faces) faces)))))))
     (t
      nil))))

(defun ns-set-foreground-at-mouse ()
  "Set the foreground color at the mouse location to ns-input-color."
  (interactive)
  (let* ((pos (mouse-position))
         (frame (car pos))
         (face (ns-face-at-pos pos)))
    (cond
     ((eq face 'cursor)
      (modify-frame-parameters frame (list (cons 'cursor-color 
                                                 ns-input-color))))
     ((not face)
      (modify-frame-parameters frame (list (cons 'foreground-color
                                                 ns-input-color))))
     (t
      (set-face-foreground face ns-input-color frame)))))

(defun ns-set-background-at-mouse ()
  "Set the background color at the mouse location to ns-input-color."
  (interactive)
  (let* ((pos (mouse-position))
         (frame (car pos))
         (face (ns-face-at-pos pos)))
    (cond
     ((eq face 'cursor)
      (modify-frame-parameters frame (list (cons 'cursor-color
                                                 ns-input-color))))
     ((not face)
      (modify-frame-parameters frame (list (cons 'background-color
                                                 ns-input-color))))
     (t
      (set-face-background face ns-input-color frame)))))



;; Misc aliases
(defalias 'x-display-mm-width 'ns-display-mm-width)
(defalias 'x-display-mm-height 'ns-display-mm-height)
(defalias 'x-display-backing-store 'ns-display-backing-store)
(defalias 'x-display-save-under 'ns-display-save-under)
(defalias 'x-display-visual-class 'ns-display-visual-class)
(defalias 'x-display-screens 'ns-display-screens)


;; Set some options to be as NS-like as possible.
(setq transient-mark-mode t
      highlight-nonselected-windows nil
      delete-selection-mode t
      search-highlight t
      query-replace-highlight t
      split-window-keep-point t
      mouse-copy-selection nil
      frame-title-format t
      icon-title-format t)
(mouse-wheel-mode 1)

;; Set up browser connectivity
(setq browse-url-browser-function 'browse-url-generic)
(cond ((eq system-type 'darwin)
       (setq browse-url-generic-program "open"))
      ;; otherwise, gnustep
      (t
       (setq browse-url-generic-program "gopen")) )

;; pull in some lisp niceties
(require 'ns-mark-nav)
(autoload 'ns-grabenv "ns-grabenv" "Get environment from your shell." t nil)



;;; Do the actual NS Windows setup here; the above code just defines
;;; functions and variables that we use now.

(setq command-line-args (ns-handle-args command-line-args))

(ns-open-connection (system-name) nil t)

(let ((services (ns-list-services)))
  (while services
    (if (eq (caar services) 'undefined)
	(ns-define-service (cdar services))
      (define-key global-map (vector (caar services))
	(ns-define-service (cdar services)))
      )
    (setq services (cdr services))))

;; For some reason this gets overwritten by ns-menu-bar.el
(load-library "mule-cmds")

(if (and (eq (get-lisp-resource nil "NXAutoLaunch") t)
         (eq (get-lisp-resource nil "HideOnAutoLaunch") t))
     (add-hook 'after-init-hook 'do-hide-emacs))

(setq frame-creation-function 'x-create-frame-with-faces)

(menu-bar-mode (if (get-lisp-resource nil "Menus") 1 -1))




;; --------------------------------------------------------------------------
;; Below here not FSF-assigned copyright, and anyway some users may not like.

;;; page-down,page-up etc. leaves cursor in same place
;;; modified (ABR) from pager.el --- windows-scroll commands
;;; Version 2.0 - 97-10-06
;;; Copyright (C) 1992-1997 Mikael Sjodin (mic@docs.uu.se)
;;; http://user.it.uu.se/~mic/emacs.shtml

(defvar pager-temporary-goal-column 0
  "Similar to temporary-goal-column but used by the pager.el functions")
;(make-variable-buffer-local 'pager-temporary-goal-column)

(defconst pager-keep-column-commands
  '(pager-row-down pager-row-up row-dn row-up
		   pager-page-down pager-page-up pg-dn pg-up)
  "Commands which when called without any other intervening command should
keep the `pager-temporary-goal-column'")

(defun pager-page-down ()
  "Like scroll-up, but moves a fixed amount of lines (fixed relative the
`window-text-height') so that pager-page-up moves back to the same line."
  (interactive)
  (if (not (pos-visible-in-window-p (point-max)))
      (pager-scroll-screen (- (window-text-height)
			   next-screen-context-lines))))

(defun pager-page-up ()
  "Like scroll-down, but moves a fixed amount of lines (fixed relative the
`window-text-height') so that pager-page-down moves back to the same line."
  (interactive)
  (if (not (pos-visible-in-window-p (point-min)))
      (pager-scroll-screen (- next-screen-context-lines
			      (window-text-height)))))

(defun pager-scroll-screen (lines)
  "Scroll screen LINES, but keep the cursors position on screen."
  (if (not (memq last-command pager-keep-column-commands))
      (setq pager-temporary-goal-column (current-column)))
  (save-excursion
    (goto-char (window-start))
    (vertical-motion lines)
    (set-window-start (selected-window) (point)))
  (vertical-motion lines)
  (move-to-column pager-temporary-goal-column))

(global-set-key "\C-v"     'pager-page-down)
(global-set-key [next]     'pager-page-down)
(global-set-key "\ev"      'pager-page-up)
(global-set-key [prior]    'pager-page-up)

;; ------------------------------

;;; mic-paren
;;; http://user.it.uu.se/~mic/emacs.shtml
;;; See ../mic-paren.el for copyright info
(require 'paren)
(load "mic-paren")
(defface paren-face-match-light
  '((((class color)) (:background "ARGBffd8eeee"))
    (t (:background "gray")))
  "Mic-paren mode face used for a matching paren."
  :group 'faces
  :group 'mic-paren-matching)
(custom-set-variables
 '(paren-sexp-mode t)
 '(paren-match-face 'paren-face-match-light))

; Redefine this to pick up mic-paren
(define-minor-mode show-paren-mode
  "Toggle Show Paren mode, implemented by mic-paren
With prefix ARG, turn Mic Paren mode on if and only if ARG is positive.
Returns the new status of Mic Paren mode (non-nil means on).

When Mic Paren mode is enabled in Emacs.app, entire s-expressions will be
highlighted by default.  To change this behavior, customize group mic-paren.

Show paren mode is implemented by mic-paren in Emacs.app.
To restore ordinary show-paren-mode, do 'M-x load-library paren'."
  :global t :group 'mic-paren
  (if show-paren-mode
    (paren-activate)
    (paren-deactivate)))


(provide 'ns-windowing)

; execute custom NS startup functions
;(load-library "ns-startup")

;;; ns-win.el ends here
