/* Image support for the NeXT/Open/GNUstep and MacOSX window system.

   Copyright (C) 1989, 1992, 1993, 1994, 2005, 2006 Free Software Foundation.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.

Originally by Carl Edman
Updated by Christian Limpach (chris@nice.ch)
OpenStep/Rhapsody port by Scott Bender (sbender@harmony-ds.com)
MacOSX/Aqua port by Christophe de Dinechin (descubes@earthlink.net)
GNUstep port and post-20 update by Adrian Robert (arobert@cogsci.ucsd.edu)

*/

#include "config.h"
#include "lisp.h"
#include "dispextern.h"
#include "nsterm.h"
#include "frame.h"

extern Lisp_Object QCfile, QCdata;

// call tracing
#if 0
int image_trace_num = 0;
#define NSTRACE(x)        fprintf(stderr, "%s:%d: [%d] " #x "\n",         \
                                __FILE__, __LINE__, ++image_trace_num)
#else
#define NSTRACE(x)
#endif


// ============================================================================
//
//   C interface.  This allows easy calling from C files.  We could just
//   compile everything as Objective-C, but that might mean slower
//   compilation and possible difficulties on some platforms..
//
// ============================================================================

void *
ns_image_from_XBM (unsigned char *bits, int width, int height)
{
  NSTRACE(ns_image_from_XBM);
  return [[EmacsImage alloc] initFromXBM: bits
                                   width: width height: height
                                    flip: YES];
}

void *
ns_image_for_XPM (int width, int height, int depth)
{
  NSTRACE(ns_image_for_XPM);
  id img = [[EmacsImage alloc] initForXPMWithDepth: depth
                                             width: width height: height];
}

void *
ns_image_from_file (Lisp_Object file)
{
  NSTRACE(ns_image_from_bitmap_file);
  return [EmacsImage allocInitFromFile: file];
}

int
ns_load_image (struct frame *f, struct image *img,
               Lisp_Object spec_file, Lisp_Object spec_data)
{
  NSTRACE(ns_load_image);

  EmacsImage *eImg;
  NSSize size;

  if (NILP (spec_data))
    {
      eImg = [EmacsImage allocInitFromFile: spec_file];
    }
  else
    {
      NSData *data = [NSData dataWithBytes: XSTRING (spec_data)->data
                                    length: SBYTES(spec_data)];
      eImg = [[EmacsImage alloc] initWithData: data];
      [eImg setPixmapData];
    }

  if (eImg == nil)
    {
      add_to_log("Unable to load image %s", img->spec, Qnil);
      return 0;
    }

  size = [eImg size];
  img->width = size.width;
  img->height = size.height;

  // 4) set img->pixmap = emacsimage
  img->pixmap = eImg;
  return 1;
}


int
ns_image_width (void *img)
{
  return [(id)img size].width;
}

int
ns_image_height (void *img)
{
  return [(id)img size].height;
}

unsigned long
ns_get_pixel(void *img, int x, int y)
{
  return [(EmacsImage *)img getPixelAtX: x Y: y];
}

void
ns_put_pixel(void *img, int x, int y, void *color)
{
  [(EmacsImage *)img setPixelAtX: x Y: y to: color];
}

void
ns_set_alpha(void *img, int x, int y, int a)
{
  [(EmacsImage *)img setAlphaAtX: x Y: y to: a];
}

int ns_get_red(void *color)
{
  return [(NSColor *)color redComponent] * 65535;
}

int ns_get_green(void *color)
{
  return [(NSColor *)color greenComponent] * 65535;
}

int ns_get_blue(void *color)
{
  return [(NSColor *)color blueComponent] * 65535;
}


// ============================================================================
//
//   Class supporting bitmaps and images of various sorts.
//
// ============================================================================

@implementation EmacsImage

static EmacsImage *ImageList = nil;

+ allocInitFromFile:(Lisp_Object)file
{
  EmacsImage *image = ImageList;
  Lisp_Object found;

  // look for an existing image of the same name
  while (image != nil &&
         [[image name] compare: [NSString stringWithUTF8String: SDATA(file)]]
             != NSOrderedSame)
    image = [image imageListNext];

  if (image != nil)
    {
      [image reference];
      return image;
    }

  /* Search bitmap-file-path for the file, if appropriate.  */
  found = x_find_image_file(file);
  if (!STRINGP(found))
    return nil;

  image = [[EmacsImage alloc] initByReferencingFile:
                     [NSString stringWithUTF8String: SDATA(found)]];

  if ([image bestRepresentationForDevice:nil] == nil)
    {
      [image release];
      return nil;
    }

  [image setName:[NSString stringWithUTF8String: SDATA(file)]];
  [image reference];
  ImageList = [image imageListSetNext:ImageList];

  return image;
}


- reference
{
  refCount++;
  return self;
}


- imageListSetNext:(id)arg
{
  imageListNext = arg;
  return self;
}


- imageListNext
{
  return imageListNext;
}


- (void)dealloc
{
  id list = ImageList;

  if (refCount > 1)
    {
      refCount--;
      return;
    }

  [stippleMask release];

  if (list == self)
    ImageList = imageListNext;
  else
    {
      while (list != nil && [list imageListNext] != self)
        list = [list imageListNext];
      [list imageListSetNext:imageListNext];
    }

  [super dealloc];
}


- initFromXBM:(unsigned char *)bits width:(int)w height:(int)h
         flip:(BOOL)flip
{
  return [self initFromSkipXBM:bits width:w height:h flip:flip length:0];
}


- initFromSkipXBM:(unsigned char *)bits width:(int)w height:(int)h
             flip:(BOOL)flip length:(int)length;
{
  NSSize s = {w, h};
  int bpr = (w + 7) / 8;
  unsigned char *planes[5];

  [self initWithSize:s];

  bmRep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes: NULL
                                    pixelsWide: w pixelsHigh: h
                                    bitsPerSample: 1 samplesPerPixel: 2
                                    hasAlpha:YES isPlanar:YES
                                    colorSpaceName: NSCalibratedBlackColorSpace
                                    bytesPerRow: bpr bitsPerPixel: 0];

  [bmRep getBitmapDataPlanes: planes];
  {
    // turn the bytes around and invert for alpha on plane 2
    int i, j;
    unsigned char *s = bits;
    unsigned char *d = planes[0];
    unsigned char *d2 = planes[1];
    unsigned char swt[16] = {0, 8, 4, 12, 2, 10, 6, 14, 1, 9, 5, 13,
                             3, 11, 7, 15};
    unsigned char c;

    for (j = 0; j < h; j++)
      for (i = 0; i < bpr; i++)
        {
          if (length)
            {
              unsigned char s1, s2;
              while (*s++ != 'x' && s < bits + length);
              if (s >= bits + length)
                {
                  [bmRep release];
                  return nil;
                }
#define hexchar(x) (isdigit (x) ? x - '0' : x - 'a' + 10)
              s1 = *s++;
              s2 = *s++;
              c = hexchar (s1) * 0x10 + hexchar (s2);
            }
          else
            c = *s++;

          *d = flip ? swt[c >> 4] | (swt[c & 0xf] << 4)
                    : c ^ 255;
          *d2++ = *d++;
        }
  }

  [self addRepresentation: bmRep];
  return self;
}



- initForXPMWithDepth: (int)depth width: (int)width height: (int)height
{
  NSSize s = {width, height};

  [self initWithSize: s];

  bmRep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes: NULL
                                  pixelsWide: width pixelsHigh: height
                                  // keep things simple for now
                                  bitsPerSample: 8 samplesPerPixel: 4 // RGB + A
                                  hasAlpha:YES isPlanar:YES
                                  colorSpaceName: NSCalibratedRGBColorSpace
                                  bytesPerRow: width bitsPerPixel: 0];

  [bmRep getBitmapDataPlanes: pixmapData];
  [self addRepresentation: bmRep];
  return self;
}


// attempt to pull out pixmap data from a BitmapImageRep;  returns NO if fails
- (void) setPixmapData
{
  NSEnumerator *reps;
  NSImageRep *rep;

  reps = [[self representations] objectEnumerator];
  while (rep = (NSImageRep *) [reps nextObject])
    {
      if ([rep respondsToSelector: @selector(getBitmapDataPlanes:)])
        {
          bmRep = (NSBitmapImageRep *) rep;
          onTiger = [bmRep respondsToSelector: @selector(colorAtX:y:)];

          if ([bmRep numberOfPlanes] >= 3)
              [bmRep getBitmapDataPlanes: pixmapData];
          break;
        }
    }
}


// note; this and next work only for image created with initForXPMWithDepth,
// initFromSkipXBM, or where setPixmapData was called successfully
- (NSColor *) getPixelAtX: (int)x Y: (int)y
{
  if (bmRep == nil)
    return 0;

  // this method is faster but won't work for bitmaps
  if (pixmapData[0] != NULL)
    {
      int loc = x + y * [self size].width;
      float r, g, b, a;

      r = pixmapData[0][loc] / 255.0;
      g = pixmapData[1][loc] / 255.0;
      b = pixmapData[2][loc] / 255.0;
      a = pixmapData[3][loc] / 255.0;

      return [NSColor colorWithCalibratedRed: r green: g blue: b alpha: a];
    }
  else if (onTiger)
    {
      return [bmRep colorAtX: x y: y];
    }
  return 0;
}

- (void) setPixelAtX: (int)x Y: (int)y to: (NSColor *)color
{
  if (bmRep == nil || color == nil)
    return;

  if (pixmapData[0] != NULL)
    {
      int loc = x + y * [self size].width;
      float r, g, b, a;

      [color getRed: &r green: &g blue: &b alpha: &a];

      pixmapData[0][loc] = r * 255;
      pixmapData[1][loc] = g * 255;
      pixmapData[2][loc] = b * 255;
      pixmapData[3][loc] = a * 255;
    }
  else if (onTiger)
    {
      [bmRep setColor: color atX: x y: y];
    }
}

- (void) setAlphaAtX: (int) x Y: (int) y to: (int) a
{
  if (bmRep == nil)
    return;

  if (pixmapData[0] != NULL)
    {
      int loc = x + y * [self size].width;

      pixmapData[3][loc] = a;
    }
  else if (onTiger)
    {
      NSColor *color = [bmRep colorAtX: x y: y];
      color = [color colorWithAlphaComponent: (a / 255.0)];
      [bmRep setColor: color atX: x y: y];
    }
}

// returns a pattern color, which is cached here
- (NSColor *)stippleMask
{
  if (stippleMask == nil)
      stippleMask = [[NSColor colorWithPatternImage: self] retain];
  return stippleMask;
}

@end
